import inspect
import warnings

def display_function_content(function):
    """ Display the code of the selected function

    Args:
        function ([function]): function to display
    """
    lines = inspect.getsource(function)
    for line in lines:
        print(line)

def notebook_autoreload():
    print('%load_ext autoreload')
    print('%autoreload 2')
    print('%aimport utils')

def add_kernel_to_notebook():
    print('kernel install --user --name=<any_name_for_kernel>')

def text_to_list():
    print('import ast')
    print('ma_list = ast.literal_eval(string)')

def pandas():
    print("pd.set_option('chained_assignment', None)")
    print("pd.set_option('display.max_rows', 500)")
    print("pd.set_option('display.max_columns', 100)")


def silent_warning():    
    warnings.warn('Removing warning from execution context.')
    warnings.filterwarnings("ignore")





import dash_daq as daq
import dash_core_components as dcc
import dash_html_components as html
from plotly import graph_objs as go


def df_to_table(df):
    """Transforms a dataFrame into an html table for Dash
    """
    return html.Table(
        # Header
        [
            html.Tr([
                html.Th(col) for col in df.columns
                     ])] +

        # Body
        [
            html.Tr(
                [
                    html.Td(df.iloc[i][col])
                    for col in df.columns
                ]
            )
            for i in range(len(df))
        ]
    )

  
def indicator(color, text, id_value):
    """returns top indicator div
    """
    return html.Div(
        [

            html.P(
                text,
                className="twelve columns indicator_text"
            ),
            html.P(
                id=id_value,
                className="indicator_value"
            ),
        ],
        className="four columns indicator",

    )


def NamedDropdown(name, **kwargs):
    return html.Div(
        style={"margin": "10px 0px"},
        children=[
            html.P(children=f"{name}", style={"margin-left": "3px"}),
            dcc.Dropdown(**kwargs),
        ],
    )


def OnOffButton(name, **kwargs):
    return html.Div(
        # style={"margin": "auto"},
        children=[
            daq.PowerButton(**kwargs)
        ],
    )


def simpleButton(name, **kwargs):
    return html.Div(
        # style={"margin": "auto"},
        children=[
            daq.StopButton(**kwargs)
        ],
    )


def circle_number(value, max_value=100):
    values = [max_value - value, value]
    colors = ['rgba(0, 0, 0, 0)', "crimson"]  # "rgb(204, 255, 255)"]
    direction = 'clockwise'
    rotation = 0 if value >= max_value / 2 else 360 / max_value * value

    data = [go.Pie(
        values=values,
        hole=.9,
        showlegend=False,
        marker={'colors': colors},
        textinfo="none",
        direction=direction,
        rotation=rotation,
    )]

    layout = go.Layout(
        margin={'l': 0, 'r': 0, 't': 0, 'b': 0},
        width=70,
        height=70,
        paper_bgcolor='rgba(0,0,0,0)',
        plot_bgcolor='rgba(0,0,0,0)',
        annotations=[
            {
                "font": {"size": 15, "color": "crimson"},
                "showarrow": False,
                "text": value,
                "align": "center",
            },
        ],
    )
    return {"data": data, "layout": layout}



import copy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from sklearn.linear_model import Lasso
from sklearn.model_selection import TimeSeriesSplit
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error


# from pandas.plotting import register_matplotlib_converters
# register_matplotlib_converters()

# # Next steps :
# - Documentation of functions
# - Create a ReadMe
# - Add a function for saving object 

class TimeSerieCrossVal:

    def __init__(self, split_number=10, regr=None, verbose=0, title='Default_title'):
        self.title = title
        self.verbose_ = verbose
        self.verbose('Init {}'.format(title), 1)
        self.score_dict = {}
        self.split_number = split_number
        self.splitter = self.initSplit(split_number)
        self.data = {'X': None, 'y': None, 'predict_by_model': None}
        self.x_split_range = None

        if regr is None:
            self.is_regr = False
        else:
            self.is_regr = True
            self.regr = copy.deepcopy(regr)

    def verbose(self, text, lvl=10):
        if self.verbose_ == -1:
            print(text)
        elif lvl <= self.verbose_:
            print(text)

    def initSplit(self, split_number):
        self.verbose('Le dataset sera séparé en {} parties.'.format(split_number), 1)
        tscv = TimeSeriesSplit(n_splits=split_number)
        return tscv

    def addRegr(self, regr):
        self.verbose('Ajout d\'un modèle de régression personnalisé.', 1)
        self.verbose(regr, 2)
        self.regr = regr
        self.is_regr = True

    def delRegr(self):
        self.verbose('Supression du modèle de régression.', 1)
        self.regr = None
        self.is_regr = False

    def addDefaultRegr(self, id_=2):
        self.verbose('Ajout d\'un modèle de régression par défaut : ({})'.format(id_), 1)
        if (id_ == 1) or (id_ == 'rf_deep'):
            self.regr = RandomForestRegressor(max_depth=4, random_state=1, n_estimators=90, min_samples_split=4)
            self.is_regr = True
        if (id_ == 2) or (id_ == 'lasso'):
            self.regr = Lasso(alpha=1)
            self.is_regr = True
        if (id_ == 3) or (id_ == 'rf'):
            self.regr = RandomForestRegressor(max_depth=2, random_state=1, n_estimators=20, min_samples_split=2)
            self.is_regr = True

    def MASE(self, diff, testing_series, prediction_series):
        n = diff.shape[0]
        d = np.abs(diff).sum() / (n - 1)
        errors = np.abs(testing_series - prediction_series)
        return errors.mean() / d

    def MAPE(self, y_train, y_pred):
        return np.max(np.abs((y_train - y_pred) / y_train)) * 100

    def displayScores(self, y_train, y_pred, name='Dataset', mase_=True, mape=False, tmp=False, cut=3):

        # Computing
        # print('DISPLAY SCORES RMSE', y_train.values, y_pred)
        rmse = np.sqrt(mean_squared_error(y_train.values, y_pred))  # metrics.roc_auc_score(y_test,pred[:,1])
        # print('rmse', rmse)
        mean_error = np.sum(np.abs((y_pred - y_train))) / len(y_pred)
        if mase_:
            mase = self.MASE(y_train - y_train.shift(-8 * 6 * 6), y_train, y_pred)
        if mape and tmp is not None:
            mape = self.MAPE(y_train + tmp.temperature_smooth, y_pred + tmp.temperature_smooth)

        # Displaying
        nb_stars = 5
        stars_line = '*' * nb_stars
        self.verbose('\n {} {} {}'.format(stars_line, name, stars_line), 2)
        self.verbose('RMSE : {}'.format(np.round(rmse, cut)), 2)
        self.verbose('Mean : {}'.format(np.round(mean_error, cut)), 2)
        if mase_:
            self.verbose('MASE : {}'.format(np.round(mase, cut)), 2)
        if mape and tmp is not None:
            self.verbose('MAPE : {}'.format(np.round(mape, cut)), 2)

        if mase_:
            return rmse, mean_error, mase
        return rmse, mean_error

    def regression(self, X_train, X_test, y_train, y_test):
        self.verbose('Starting regression', 2)
        regressor = copy.deepcopy(self.regr)
        regressor.fit(X_train, y_train)
        y_pred_train = regressor.predict(X_train)
        y_pred_test = regressor.predict(X_test)
        self.verbose(regressor, 3)
        self.verbose('Ending regression', 2)
        return (y_pred_train, y_pred_test, regressor)

    def fit(self, X, y, compute_mase_=True):
        if not self.is_regr:
            self.addDefaultRegr(2)
        self.data['X'] = X.copy()
        self.data['y'] = y.copy()
        self.verbose('Dimension X : {}'.format(X.shape), 1)
        self.verbose('Dimension y : {}'.format(y.shape), 1)
        self.score_dict = {}
        self.x_split_range = None

        id_model = 0
        for _, test_index in self.splitter.split(X):
            self.verbose('Iteration : {}/{}'.format(id_model + 1, self.split_number), 1)

            # Separating train/test datasets
            X_train_cv, X_test_cv = X.iloc[:test_index[0]], X.iloc[test_index[0]:test_index[-1]]
            y_train_cv, y_test_cv = y.iloc[:test_index[0]], y.iloc[test_index[0]:test_index[-1]]

            # Printing information about split£
            self.verbose('Dimensions de X_train {}'.format(X_train_cv.shape), 2)
            self.verbose('Dimensions de X_test  {}'.format(X_test_cv.shape), 2)
            self.verbose('Dimensions de y_train {}'.format(y_train_cv.shape), 2)
            self.verbose('Dimensions de y_test  {}'.format(y_test_cv.shape), 2)

            # Calculating 
            regr_result = self.regression(X_train_cv, X_test_cv, y_train_cv, y_test_cv)
            y_pred_train = regr_result[0]
            y_pred_test = regr_result[1]
            regressor = copy.deepcopy(regr_result[2])

            # Storing results
            ans = {'y_pred_train': y_pred_train,
                   'y_pred_test': y_pred_test,
                   'train_score': self.displayScores(y_train_cv, y_pred_train, 'TRAIN', mase_=compute_mase_),
                   'test_score': self.displayScores(y_test_cv, y_pred_test, 'TEST', mase_=compute_mase_),
                   # , mape=True, tmp=test_set)],
                   'regr': regressor}
            self.score_dict[id_model] = ans
            id_model += 1

        self.x_split_range = [value['y_pred_train'].shape[0] for key, value in self.score_dict.items()]

    def plotScores(self, test=True, train=False, rmse_=True, mean_=False):
        # %matplotlib inline
        _ = plt.subplots(figsize=(20, 5))

        if test:
            if rmse_:
                _ = plt.plot(self.x_split_range, [value['test_score'][0] for key, value in self.score_dict.items()],
                             marker='o', label='RMSE_test')
            if mean_:
                _ = plt.plot(self.x_split_range, [value['test_score'][1] for key, value in self.score_dict.items()],
                             marker='o', label='MEAN_test')
            # Test qui marche mais foireux avec le 0...
            if len(self.score_dict[0]['test_score']) > 2:
                _ = plt.plot(self.x_split_range, [value['test_score'][2] for key, value in self.score_dict.items()],
                             marker='o', label='MASE_test')

        if train:
            if rmse_:
                _ = plt.plot(self.x_split_range, [value['train_score'][0] for key, value in self.score_dict.items()],
                             marker='o', label='RMSE_train')
            if mean_:
                _ = plt.plot(self.x_split_range, [value['train_score'][1] for key, value in self.score_dict.items()],
                             marker='o', label='MEAN_train')
            # Test qui marche mais foireux avec le 0...
            if len(self.score_dict[0]['test_score']) > 2:
                _ = plt.plot(self.x_split_range, [value['train_score'][2] for key, value in self.score_dict.items()],
                             marker='o', label='MASE_train')

        _ = plt.xlabel('Nombre de lignes d\'apprentissage')
        _ = plt.ylabel('Valeur de l\'erreur')
        _ = plt.title('Erreur en CV Time Series')
        _ = plt.legend()
        _ = plt.show()

    def displayMeanMetrics(self, mase_=True, cut=4):
        print('RMSE MOYEN :',
              np.round(np.mean([value['test_score'][0] for key, value in self.score_dict.items()]), cut))
        print('MEAN MOYEN :',
              np.round(np.mean([value['test_score'][1] for key, value in self.score_dict.items()]), cut))
        if mase_:
            print('MASE MOYEN :',
                  np.round(np.mean([value['test_score'][2] for key, value in self.score_dict.items()]), cut))

    def getMeanMetrics(self, mase_=True, cut=4):
        rmse = np.round(np.mean([value['test_score'][0] for key, value in self.score_dict.items()]), cut)
        mean = np.round(np.mean([value['test_score'][1] for key, value in self.score_dict.items()]), cut)
        if mase_:
            mase = np.round(np.mean([value['test_score'][2] for key, value in self.score_dict.items()]), cut)
            return rmse, mean, mase
        return rmse, mean

    def createModelPredic(self, return_new=False):

        new_df = self.data['X'].copy()
        index_name = new_df.index.name

        new_df['_model_'] = -1
        new_df['_datetime_'] = new_df.index
        new_df = new_df.reset_index(drop=True)

        self.verbose('Création de la colonne _model_', 1)
        c = 0
        last = 0
        for i in range(len(self.x_split_range)):
            first, last = last, self.x_split_range[i]
            new_df.loc[(new_df.index >= first) & (new_df.index <= last), '_model_'] = c
            c += 1

        self.verbose('Création de la colonne _predic_', 1)
        new_df.loc[new_df['_model_'] == -1, '_model_'] = c
        new_df.set_index('_datetime_', inplace=True)
        new_df.index.name = index_name
        for i in new_df['_model_'].unique():
            if i > 0:
                new_df.loc[(new_df['_model_'] == i), '_predic_'] = self.score_dict[i - 1]['regr'].predict(
                    self.data['X'].copy().loc[new_df['_model_'] == i])

        self.data['predict_by_model'] = new_df[['_model_', '_predic_']]
        if return_new:
            return new_df

    def plotPredictions(self, nb=-1, y=True):
        if (nb not in self.score_dict.keys()) and nb != -1:
            max_ = max(self.score_dict.keys())
            print('No model named {}. Max is {}'.format(nb, max_))

        new_df = pd.concat([self.data['X'], self.data['predict_by_model']], axis=1)
        new_df['_target_'] = self.data['y']
        new_df['_datetime_'] = new_df.index
        _ = plt.subplots(figsize=(20, 5))

        # Plotting all models results per chunk
        if nb == -1:
            for i in new_df['_model_'].unique():
                if i != -1:
                    abscisses = new_df[new_df['_model_'] == i]['_datetime_']
                    ordonnees = new_df[new_df['_model_'] == i]['_predic_']
                    _ = plt.plot(abscisses, ordonnees, label=i)
                    _ = plt.title('Results of all CV-models')

        # Plotting results of one model on all data
        else:
            # Maybe use the predict function 
            mid_index = self.x_split_range[nb]
            regressor = copy.deepcopy(self.score_dict[nb]['regr'])
            time_value = new_df['_datetime_']
            to_predict = self.data['X'].copy()
            predicted_value = regressor.predict(to_predict)

            _ = plt.plot(time_value, predicted_value, label='test')
            _ = plt.plot(time_value.iloc[:mid_index], predicted_value[:mid_index], label='train')
            _ = plt.title('Model {}'.format(nb))

        if y:
            _ = plt.plot(new_df['_datetime_'], new_df['_target_'], c='grey', alpha=0.3)

        # Finally add ledeng to the plot
        _ = plt.legend()
        # if nb!=-1:
        #    return self.score_dict[nb]

    def getYValues(self, nb_model):

        # Prediction values
        y_pred_train = self.score_dict[nb_model]['y_pred_train']
        y_pred_test = self.score_dict[nb_model]['y_pred_test']

        # Real values
        y = self.data['y']
        mid_index = self.x_split_range[nb_model]
        y_train = y.iloc[:mid_index]
        y_test = y.iloc[mid_index:mid_index + len(y_pred_test)]

        return y_train, y_test, y_pred_train, y_pred_test

    def getXValues(self, nb_model):
        X = self.data['X']
        mid_index = self.x_split_range[nb_model]
        end_index = mid_index + len(self.score_dict[nb_model]['y_pred_test'])
        X_train = X.iloc[:mid_index]
        X_test = X.iloc[mid_index:end_index]
        return X_train, X_test

    def getOriginalData(self):
        X = self.data['X']
        y = self.data['y']
        return X, y

    def saveObject(self):
        result = {'title': self.title,
                  'nb_split': self.split_number,
                  'score_dict': self.score_dict,
                  'x_split_range': self.x_split_range}
        return result

    def restaure_from_dict(self, result, X, y):
        self.data['X'] = X
        self.data['y'] = y
        self.title = result['title']
        self.split_number = result['nb_split']
        self.score_dict = result['score_dict']
        self.x_split_range = result['x_split_range']
        self.initSplit(self.split_number)
        self.createModelPredic()



import os
import json
import pickle
import requests


def parse_json(path):
    """Given a json_file =path loading it and sending a dictionary back"""
    with open(path) as f:
        data = json.load(f)
    return data


def url_to_json(url):
    """ Given a web url with a json file, returns a dictionary"""
    r = requests.get(url)
    json_file = r.json()
    return json_file


def save_json(file_name, data, path='./'):
    """Saving as a json file in one line"""
    file_path = path + file_name
    with open(file_path, 'w') as fp:
        json.dump(data, fp)


def read_file(file_name='data.txt'):
    """ Example of file name 'data.txt' """
    with open(file_name, "r") as fichier:
        file = fichier.read()
    return file


def get_token(path):
    with open(path, 'r') as file:
        api_key = file.read()
    return api_key


def load_pickle(file_name):
    PICKLE_PATH = parse_json('../src/params.json')['PICKLE_PATH']
    file_path = PICKLE_PATH + file_name
    with open(file_path, 'rb') as pfile:
        my_pickle = pickle.load(pfile)
    return my_pickle


def save_pickle(object_, file_name):
    PICKLE_PATH = parse_json('../src/params.json')['PICKLE_PATH']
    file_path = PICKLE_PATH + file_name
    with open(file_path, 'wb') as pfile:
        pickle.dump(object_, pfile, protocol=pickle.HIGHEST_PROTOCOL)


def list_pickle():
    PICKLE_PATH = parse_json('../src/params.json')['PICKLE_PATH']
    file_list = os.listdir(PICKLE_PATH)
    pickle_list = [i for i in file_list if '.p' in i]
    print(pickle_list)


# YAML
# import yaml
# import pathlib

# def load_config(config_filename):
#     with open(pathlib.Path(__file__).parent / config_filename) as stream:
#         data_loaded = yaml.safe_load(stream)
#     return data_loaded

# with open("drift_config.yaml", 'r') as stream:
#     data_loaded = yaml.safe_load(stream)

# print(data_loaded)

# error_limit:
#   MAPE_14H: 5
#   MAPE_15H: 10
#   MAPE_16H: 10
#   MAPE_17H: 10
#   MAPE_18H: 10
#   MAPE_19H: 10
#   MAPE_20H: 10
# avert_data:
#   nb_days_analyse: 7
#   nb_days_alert: 2
#   color: FFD700
#   titre: Dérive du modèle - Alerte
#   message: Le modèle commence à dériver, un ré-entrainement pourrait survenir prochainement
# retrain_data:
#   nb_days_analyse: 7
#   nb_days_alert: 14
#   color: 850606
#   titre: Dérive du modèle - Ré-entrainement
#   message: Le modèle dérive depuis 10 jours, un ré-entrainement vient d'être enclenché
# webhoook:
#   url: https://mousquetaires.webhook.office.com/webhookb2/a2e753ef-4c43-4ecc-a111-dde5c8767ed8@ebaeb74c-eeae-40f8-b755-10b0e2bf7528/IncomingWebhook/5c323e2b2e254db091e31464d032336e/49c2927e-6269-4317-9d68-d2d1057b07fb


# Making imports
import numpy as np

# Constant values 
ROUND = 8


def is_crossing(x1, y1, x2, y2, x3, y3, x4, y4):
    a1, b1 = get_a_b_2_pts(x1, y1, x2, y2)
    a2, b2 = get_a_b_2_pts(x3, y3, x4, y4)

    # Cas particuliers
    if a1 == a2:
        if b1 == b2:
            # print("Droites identiques")
            return False
        # print("Droites parallèles")
        return False

    # Cas général
    if (a1 == 'vert') & (a2 == 'vert'):
        # print('Vert both')
        return False
    elif a1 == 'vert':
        # print('Vert a1')
        x_inter = b1
        y_inter = a2 * x_inter + b2
        if ((y_inter - y1) * (y_inter - y2)) < 0:
            return True
        return False

    #    elif a2 == 'vert':
    #        # print('Vert a2')
    #        x_inter = b2
    #        y_inter = a1*x_inter+b1
    #        if ((y_inter - y1) * (y_inter - y2)) < 0:
    #            return True
    #        return False

    else:
        x_inter = (b2 - b1) / (a1 - a2)

    if ((x_inter - x1) * (x_inter - x2)) < 0 and (x_inter - x3) * (x_inter - x4) < 0:
        return True
    return False


def get_a_b_hauteur_pt(a_dr, b_dr, x1, y1):
    a = -1 / a_dr
    b = y1 - a * x1
    return np.round(a, ROUND), np.round(b, ROUND)


def get_a_b_bissectrice(x1, y1, x2, y2, x3, y3):
    x_m = (x1 + x2) / 2
    y_m = (y1 + y2) / 2
    a, b = get_a_b_2_pts(x_m, y_m, x3, y3)
    return np.round(a, ROUND), np.round(b, ROUND)


def get_a_b_2_pts(x1, y1, x2, y2):
    if x2 == x1:
        # print('Droite verticale')
        return 'vert', x1

    a = (y2 - y1) / (x2 - x1)
    b = y1 - a * x1
    return np.round(a, ROUND), np.round(b, ROUND)



import time as __time
import datetime as __datetime
import numpy as __np


def format_to_numeric_timestamp(time_var):
    # print('Start time_conversion with : {0}, of type : {1}'.format(time_var, type(time_var)))
    # TRY TO DEAL WITH : <class 'pandas._libs.tslib.Timestamp'>
    # #(when a date in a panda index) - use str() on it otherwise
    if type(time_var) == __datetime.datetime:
        time_var = str(time_var)

    if type(time_var) == float or type(time_var) == int:
        # print('Float or int : almost the good format')
        if time_var > 10000000000:
            # print('timestamp sent in microseconds : converted in miliseconds')
            return time_var / 1000
        # print('timestamp sent in milisenconds : OK')
        return time_var

    elif type(time_var) == str:
        format_1 = '%Y-%m-%d %H:%M:%S'
        format_2 = "%Y-%m-%dT%H:%M:%S.%f%z"
        format_3 = "%Y-%m-%dT%H:%M:%S.%fZ"
        format_4 = "%Y-%m-%dT%H:%M:%S.%f+0000"
        format_5 = "%Y-%m-%dT%H:%M:%S.%f+0000"
        format_6 = "%Y-%m-%dT%H:%M:%S"

        for time_format in [format_1, format_2, format_3, format_4, format_5, format_6]:
            try:
                # numeric_timestamp = time.strftime(time_format,time.gmtime(time_var))
                numeric_timestamp = __time.mktime(__datetime.datetime.strptime(time_var, time_format).timetuple())
                # print('String corresponds to format : {}'.format(time_format))
                return int(numeric_timestamp)
            except:
                # print('String : {1} doesn\'t match with format {0}'.format(time_format, time_var))
                pass
        # print('String sent : {} doesn\'t match with any registered format'.format(time_var))

        return time_var

    else:
        print('Type of time variable, is not a string neither a float or int, but {}'.format(type(time_var)))
        print('Thus conversion was not realized, return : {}'.format(time_var))
        return time_var


def format_to_df_time(format_):
    """Takes any format of date and transform it to dataFrame timestamp format : %Y-%m-%d %H:%M:%S"""
    if not format_:
        return None
    if format_ == __np.NaN:
        return None
    numeric_timestamp = format_to_numeric_timestamp(format_)
    dt = __datetime.datetime.utcfromtimestamp(numeric_timestamp)
    string = dt.strftime('%Y-%m-%d %H:%M:%S')
    return string


def format_to_date_time(format_, as_string=False):
    """Takes any format of date and transform it to datetime format.
    Arg : as_string : set to true, transfroms it into a classic human readable format"""
    if not format_:
        return None
    numeric_timestamp = format_to_numeric_timestamp(format_)
    datetime_ts = __datetime.datetime.utcfromtimestamp(numeric_timestamp)
    if as_string:
        return str(datetime_ts)
    return datetime_ts


def format_to_cloud_string(format_):
    """Takes any format of date and transform it to cloud timestamp format : %Y-%m-%dT%H:%M:%S.000Z"""
    if not format_:
        return None
    numeric_timestamp = format_to_numeric_timestamp(format_)
    dt = __datetime.datetime.utcfromtimestamp(numeric_timestamp)
    string = dt.strftime("%Y-%m-%dT%H:%M:%S.000Z")
    return string


def time_between_two_dates(start, end):
    start_datetime = format_to_date_time(start)
    end_datetime = format_to_date_time(end)

    nb_days = end_datetime - start_datetime
    nb_days_str = str(nb_days)

    return nb_days_str


def current_time_string():
    """Description : renvoie l'heure actuelle UTC au format AAAAMMJJ_HHMMSS 
    inputs :
        - None
    outputs:
        - string : heure au format indiqué ci-dessus
    """
    return str(__datetime.datetime.now()).replace(' ', '_').replace(':', '').replace('-', '')[:15]



import numpy as np
import pandas as pd
import chart_studio
import matplotlib.pyplot as plt
from plotly import graph_objs as go
import sklearn
import seaborn as sns
import itertools

def plot_results(x_values, y_init_concat, y_pred_concat, zoomable=False):
    # Use of classic motplotlib
    if not zoomable:
        plt.subplots(figsize=(10, 4))
        plt.plot(x_values, y_init_concat, label='truth')  # , marker='o')
        plt.plot(x_values, y_pred_concat, label='prediction')  # , marker='o')
        plt.legend()
        plt.show()

    # Use of Plotly
    else:
        temp_real = go.Scatter(
            x=pd.Series(x_values),
            y=pd.Series(y_init_concat))

        temp_predict = go.Scatter(
            x=pd.Series(x_values),
            y=pd.Series(y_pred_concat))

        data = [temp_real, temp_predict]
        chart_studio.plotly.iplot(data, world_readable=True)

    return None


def regression(regr, X_train, X_test, y_train, y_test):
    print('Train :', X_train.shape, 'Test : ', X_test.shape, 'Total :', X_train.shape[0] + X_test.shape[0])
    regr.fit(X_train, y_train)

    y_pred_train = regr.predict(X_train)
    train_score = display_scores(y_train, y_pred_train, 'TRAIN')

    y_pred_test = regr.predict(X_test)
    test_score = display_scores(y_test, y_pred_test, 'TEST')  # , mape=True, tmp=test_set)

    return y_pred_train, y_pred_test, train_score, test_score, regr


def MASE_score(testing_series, prediction_series):
    """
    Computes the MEAN-ABSOLUTE SCALED ERROR forcast error for univariate time series prediction. 
    See "Another look at measures of forecast accuracy", Rob J Hyndman
    parameters:
        training_series: the series used to train the model, 1d numpy array
        testing_series: the test series to predict, 1d numpy array or float
        prediction_series: the prediction of testing_series, 1d numpy array (same size as testing_series) or float
        absolute: "squares" to use sum of squares and root the result, "absolute" to use absolute values.
    
    """
    diff = testing_series - prediction_series
    n = diff.shape[0]
    print('%' * 10, n)
    d = np.abs(diff).sum() / (n - 1)

    errors = np.abs(testing_series - prediction_series)
    return errors.mean() / d


def MAPE(y_true, y_pred):
    return np.max(np.abs((y_true - y_pred) / y_true)) * 100


def display_scores(y, y_pred, name='Dataset', mape=False, tmp=False, cut=3):
    nb_stars = 5
    print('\n', '*' * nb_stars, name, '*' * nb_stars)

    # RMSE
    rmse = np.sqrt(sklearn.metrics.mean_squared_error(y, y_pred))  # metrics.roc_auc_score(y_test,pred[:,1])
    print('RMSE', np.round(rmse, cut))

    # Mean Error
    mean_error = np.sum(np.abs((y_pred - y))) / len(y_pred)
    print('Mean', np.round(mean_error, cut))

    # MASE
    mase = MAPE(y - y.shift(-8 * 6 * 6), y, y_pred)
    print('MASE', np.round(mase, cut))

    # MAPE
    if mape and tmp is not None:
        mape = MAPE(y + tmp.temperature_smooth, y_pred + tmp.temperature_smooth)
        print('MAPE', np.round(mape, cut))

    return rmse, mean_error, mase


# def select_regr_model(mdl=0):
#     if mdl == 0:
#         return SVR()
#     if mdl == 1:
#         return SVR(gamma='scale', C=1.0, epsilon=0.2)
#     if mdl == 2:
#         return RandomForestRegressor(max_depth=6, random_state=1, n_estimators=60, min_samples_split=4)
#     if mdl == 3:
#         return Ridge(alpha=1.0)
#     if mdl == 4:
#         return ElasticNet()
#     return Lasso


def split_train_test(df):
    # Select temporal range
    start_date = pd.to_datetime('20170115')
    end_date = pd.to_datetime('20180501')

    # Separating dataset into two
    train_set = df[(df.datetime > start_date) &
                   (df.datetime < end_date)].copy()

    test_set = df[df.datetime > end_date].copy()
    return train_set, test_set


# i = 1
# _ = plt.figure(figsize=(20, 10))
# for type_ in df_pred.type.unique():
#     _ = plt.subplot(3, 3, i)
#     df_pred.loc[df_pred['type'] == type_, 'y_reel'].hist(label='reel')
#     df_pred.loc[df_pred['type'] == type_, 'y_pred'].hist(label='pred')
#     _ = plt.title(type_)
#     i += 1
# _ = plt.legend()
# _ = plt.show()


# # Plotting it
# _ = sns.countplot(df_train['isFraud'])
# _ = plt.title('Répartion des fraudes')
# _ = plt.show()


# # Visualisation of missing values
# _ = plt.figure(figsize=(20, 8))
# _ = sns.barplot(y=df_train.isnull().sum().sort_values(ascending=False),
#                 x=df_train.isnull().sum().sort_values(ascending=False).index)
# _ = plt.title("Counts of missing value in df_train", size=20)
# _ = plt.xticks(fontsize=7)
# _ = plt.xticks(rotation=90)
# _ = plt.show()


def plot_hist(train, colname):
    """f.plot_hist(df_train, 'TransactionAmt')
    """
    _train_0 = train[train['isFraud'] == 0].reset_index(drop=False)
    _train_1 = train[train['isFraud'] == 1].reset_index(drop=False)
    _ = plt.figure(figsize=(20, 7))  # facecolor='w')
    ax = sns.kdeplot(_train_0[colname], color='b', alpha=0.4, shade=True)
    ax_2 = ax.twinx()
    _ = sns.kdeplot(_train_1[colname], color='r', alpha=0.4, shade=True, ax=ax_2)
    _ = ax.set_title(colname)
    _ = plt.show()


def plot_distribution_amount(df):
    # f.plot_distribution_amount(df_train)

    df1 = df[df['isFraud'] == 0]
    not_fraud = df1['TransactionAmt'].apply(np.log)  # we will apply log transformation to get better visualization

    df2 = df[df['isFraud'] == 1]
    fraud = df2['TransactionAmt'].apply(np.log)  # we will apply log transformation to get better visualization

    _ = plt.figure(figsize=(20, 7))
    _ = sns.distplot(a=not_fraud, label='Not Fraud')
    _ = sns.distplot(a=fraud, label='Fraud')
    _ = plt.legend()
    _ = plt.show()


def plot_confusion_matrix(cm, classes,
                          normalize=False,
                          title='Confusion matrix',
                          cmap=plt.cm.Blues):
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, cm[i, j],
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')


def visualize_numeric_columns(df, num_cols, x=3, y=3):
    _ = plt.figure(figsize=(20, 15))
    j = 1
    for i in num_cols:
        _ = plt.subplot(x, y, j)
        _ = sns.distplot(a=df[i])
        j = j + 1
    _ = plt.show()


# def plot_auc(reg, X_test, y_test, model_name='Logistic Regression'):
#     # Calculating output porbas and ROC Metrics
#     y_pred_prob = reg.predict_proba(X_test)[:, 1]
#     fpr, tpr, thresholds = roc_curve(y_test, y_pred_prob)
#
#     # PLotting results
#     plt.plot([0, 1], [0, 1], 'k--')
#     plt.plot(fpr, tpr, label=model_name)
#     plt.xlabel('False Positive Rate')
#     plt.ylabel('True Positive Rate')  # Is also the Recall
#     plt.title('{} ROC Curve'.format(model_name))
#     plt.show()
#
#     # Caluclation AUC Score
#     auc_score = roc_auc_score(y_test, y_pred_prob)
#     print('The AUC Score is {}'.format(auc_score))
#
#     # Calculating CV-AUC Score
#     # cv_auc_scores = cross_val_score(reg,X,y,cv=5,scoring='roc_auc')
#     # print("AUC scores computed using 5-fold cross-validation: \n{}".format(cv_auc_scores))



# For postgres information
# https://www.postgresqltutorial.com/

# Imports
import sys
import time
import psycopg2
import requests
import numpy as np
import pandas as pd

  

class pgSqlManipulation():
    """ This class aims to facilitate connections to postrgre platforms
    """
    
    def __init__(self, env=None):
        
        # Connection to Postgre
        self.get_params()
        self.instanciate_connection()
        

    #####################################################################################
    #                              SQL Part (POSTGRES)
    #####################################################################################


    def get_params(self, env=None):
        """ Used to get the environment connection to postgres database
        """
        self.host = '172.17.0.1'
        self.db = 'postgres'
        self.user = 'admin'
        self.pwd = 'password'


    def instanciate_connection(self):
        self.conn = psycopg2.connect(host=self.host, database=self.db, user=self.user, password=self.pwd)
        self.cur = self.conn.cursor()


    def sql(self, sql_request):
        """ Execute an sql request on the specified dev environment
        """
        try:
            self.cur.execute(sql_request)
            self.conn.commit()
            print("Request executed: {}...".format(sql_request[:30]))
        except:
            self.reset_connection()
            raise


    def drop_table(self,table, cascade=False):
        request = 'drop table {};'.format(table)
        if cascade:
            print('WARNING: print CASCADE executed.')
            request = 'drop table {} cascade;'.format(table)
        

        rep = self.sql(request)
        # print(rep.status_code)


    def extract_table_from_request(self, request):
        post_from = request.lower().split('from')[1]
        table_name = post_from.split()[0]
        return table_name
        

    def sql_to_df(self, sql_request, with_col=False, index=None):
        """Converts an SQL request into a dataFrame
        - with_col : If True, send an other requests to get the column names of the table and set it to the DataFrame
        - index : if specified, set the selected column as index of the DataFrame 
        """
        self.cur.execute(sql_request)
        rows = self.cur.fetchall()
        df = pd.DataFrame(rows)

        if not with_col:
            return df

        table_name = self.extract_table_from_request(sql_request)
        columns = self.get_column_names(table_name)
        df.columns = columns

        return df if index is None else df.set_index(index)
    

    def get_df_with_id_as_index(self, sql_request):
        """Use the sql_to_df functions with specific parameters (Get column names = True, index=id)
        - Warning : will not work with materialized view (only table)
        """
        return self.sql_to_df(sql_request, with_col=True, index='id')


    def get_column_names(self, table_name):
        """ This function gets the column names of a table reqeusting its schema
        - Warning : will not work with materialized view (only classical table)
        """
        sql = """select column_name
        from INFORMATION_SCHEMA.COLUMNS
        where TABLE_NAME='{}';""".format(table_name)
        self.cur.execute(sql)
        columns = self.cur.fetchall()
        return [col[0] for col in columns]


    def reset_connection(self):
        """In case of problem in a request, there might be a connection issue.
            This function reset the connection and allows to keep the object alive.
        """
        self.conn.close()
        self.conn = psycopg2.connect(
            host=self.host,
            database=self.db,
            user=self.user,
            password=self.pwd
            )
        self.cur = self.conn.cursor()


    def create_sql_request_header(self, dataframe, table_name):
        """ (1/2) Used in the process of inserting data into a pocoto database.
            This is the creation of the header of the request
            Part 2 is 'create_sql_request_row' function
        """
        s = "INSERT INTO "
        s += table_name + "("
        for col in dataframe.columns:
            s+=col.lower() + ","
        s = s[:-1] + ") VALUES ("
        for i in range(len(dataframe.columns)):
            s+= "%s,"
        s = s[:-1] + ");"
        return s


    def create_sql_request_row(self, dataframe_row):
        """ (2/2) Used in the process of inserting data into a pocoto database.
            This is the creation of the content of the request
            Part 1 is 'create_sql_request_header' function
        """
        y = []
        for i in range(len(dataframe_row)):
            if pd.isnull(dataframe_row[i]) or dataframe_row[i] == "":
                y.append(None)
            else:
                val = dataframe_row[i]

                y.append(val)
        return y


    def df_2_table(self, df, table_name, create_table=False):
        """ Used to insert data into a pocoto DataBase, given a DataFrame and a table_name.
            NB : The table must have the columns of the DataFrame as fields.
        """
        if create_table:
            self.create_table_from_df(df, table_name)

        sql_request = self.create_sql_request_header(df, table_name)
        print(sql_request)

        for index, rows in df.iterrows():
            my_sql_row = self.create_sql_request_row(rows)
            # print(my_sql_row)
            self.cur.execute(sql_request,my_sql_row)
        self.conn.commit()
        print('> {} : dataframe correctly inserted.'.format(table_name))





    def create_table_from_df(self, df, table_name):
        cols = df.dtypes.apply(str)

        #########################################################
        # !!!!! Fulfill it. Only done for so far needeed types
        #########################################################
        cols = cols.replace('object', 'text')
        cols = cols.replace('int64', 'int8')
        cols = cols.replace('float64', 'float8')
        #########################################################
        cols = cols.to_dict()
        
        structure_table = ''
        for k, v in cols.items():
            structure_table += k + ' ' + v + ','

        structure_table = 'CREATE TABLE ' + table_name + '(' + structure_table[:-1] + ');'
        self.sql(structure_table)

    

    def get_tables(self):
        request = "SELECT * FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog' AND schemaname != 'information_schema';"
        return self.sql_to_df(request)

    def test(self):
        print('Salut les amis')
        


from . import dash_functions
from . import database_test
from . import disk_interaction
from . import general
from . import geometry
from . import ml
from . import ml_viz
from . import secundary
from . import serie_gestion
from . import slack
from . import time_format
from . import ts_analysis



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='uthibs',                   # How you named your package folder (MyLib)
    packages=['uthibs'],             # Chose the same as "name"
    version='0.0.2',                 # Start with a small number and increase it with every change you make
    author='Thibaud Lamothe',        # Type in your name
    author_email='hello@etomal.com', # Type in your E-Mail
    license='MIT',                   # https://help.github.com/articles/licensing-a-repository
    description='Re-usable functions from my different projects',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/ThibaudLamothe/uthibs',
    download_url='https://github.com/ThibaudLamothe/uthibs/archive/refs/tags/v.0.0.1.tar.gz',
    keywords=['FUNCTIONS', 'UTILS', 'GENERIC'],
    install_requires=[
        'chart_studio==1.1.0',
        'dash==2.0.0',
        'dash-core-components==2.0.0',
        'dash-html-components==2.0.0',
        'matplotlib==3.4.3',
        'numpy==1.21.3',
        'pandas==1.3.4',
        'Pillow==8.4.0',
        'plotly==5.3.1',
        'psycopg2-binary==2.9.1',
        'requests==2.26.0',
        'ruptures==1.1.5',
        'scikit-learn==1.0.1',
        'scipy==1.7.1',
        'seaborn==0.11.2',
        'slacker==0.14.0',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',              # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',              # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',       # Again, pick a license
        'Programming Language :: Python :: 3.7',        # Specify which pyhton versions that you want to support
    ],
)
